#include "fit.h"
#include "usefull.h"
#include "message.h"
#include "spline.h"
#include "vector.h"
#include "curve.h"
#include "pxl-outline.h"
#include "epsilon-equal.h"
#include "xmem.h"
#include <math.h>
#include <limits.h>
#include <float.h>
#include <string.h>
#include <assert.h>


typedef struct index_list
{
  unsigned *data;
  unsigned length;
} index_list_type;


#define GET_INDEX(i_l, n)  ((i_l).data[n])
#define INDEX_LIST_LENGTH(i_l)  ((i_l).length)
#define GET_LAST_INDEX(i_l)  ((i_l).data[INDEX_LIST_LENGTH (i_l) - 1])

static void append_index (index_list_type *, unsigned);
static void free_index_list (index_list_type *);
static index_list_type new_index_list (void);
static void remove_adjacent_corners (index_list_type *, unsigned, bool);
static void align (spline_list_type *, fitting_opts_type *);
static void change_bad_lines (spline_list_type *,
  fitting_opts_type *);
static void filter (curve_type, fitting_opts_type *);
static real filter_angle (vector_type, vector_type);
static void find_curve_vectors
  (unsigned, curve_type, unsigned, vector_type *, vector_type *, unsigned *);
static unsigned find_subdivision (curve_type, unsigned initial,
  fitting_opts_type *);
static void find_vectors
  (unsigned, pixel_outline_type, vector_type *, vector_type *, unsigned);
static index_list_type find_corners (pixel_outline_type,
  fitting_opts_type *);
static real find_error (curve_type, spline_type, unsigned *);
static vector_type find_half_tangent (curve_type, bool start, unsigned *, unsigned);
static void find_tangent (curve_type, bool, bool, unsigned);
static spline_type fit_one_spline (curve_type);
static spline_list_type *fit_curve (curve_type,
  fitting_opts_type *);
static spline_list_type fit_curve_list (curve_list_type,
  fitting_opts_type *);
static spline_list_type *fit_with_least_squares (curve_type,
  fitting_opts_type *);
static spline_list_type *fit_with_line (curve_type);
static void remove_knee_points (curve_type, bool);
static bool reparameterize (curve_type, spline_type);
static void set_initial_parameter_values (curve_type);
static bool spline_linear_enough (spline_type *, curve_type,
  fitting_opts_type *);
static curve_list_array_type split_at_corners (
  pixel_outline_list_type, fitting_opts_type *);
static bool test_subdivision_point (curve_type, unsigned,
  vector_type *, fitting_opts_type *);
static coordinate_type real_to_int_coord (real_coordinate_type);
static real distance (real_coordinate_type, real_coordinate_type);

fitting_opts_type new_fitting_opts (void)
{
  fitting_opts_type fitting_opts;
  fitting_opts.align_threshold = (real) 0.5;
  fitting_opts.bgColor = NULL;
  fitting_opts.color_count = 0;
  fitting_opts.corner_always_threshold = (real) 60.0;
  fitting_opts.corner_surround = 3;
  fitting_opts.corner_threshold = (real) 100.0;
  fitting_opts.error_threshold = (real) 2.3;
  fitting_opts.filter_alternative_surround = 5;
  fitting_opts.filter_epsilon = (real) 20.0;
  fitting_opts.filter_iteration_count = 4;
  fitting_opts.filter_percent = (real) 0.33;
  fitting_opts.filter_secondary_surround = 3;
  fitting_opts.filter_surround = 2;
  fitting_opts.line_reversion_threshold = (real) .01;
  fitting_opts.line_threshold = (real) 1.0;
  fitting_opts.remove_adj_corners = false;
  fitting_opts.reparameterize_improvement = (real) .10;
  fitting_opts.reparameterize_threshold = (real) 30.0;
  fitting_opts.subdivide_search = (real) .2;
  fitting_opts.subdivide_surround = 4;
  fitting_opts.subdivide_threshold = (real) .03;
  fitting_opts.tangent_surround = 3;

  return (fitting_opts);
}


spline_list_array_type
fitted_splines (pixel_outline_list_type pixel_outline_list,
  fitting_opts_type *fitting_opts)
{
  unsigned this_list;
  unsigned total = 0;
  spline_list_array_type char_splines = new_spline_list_array ();
  curve_list_array_type curve_array = split_at_corners (pixel_outline_list, fitting_opts);

  for (this_list = 0; this_list < CURVE_LIST_ARRAY_LENGTH (curve_array);
       this_list++)
    {
      spline_list_type curve_list_splines;
      curve_list_type curves = CURVE_LIST_ARRAY_ELT (curve_array, this_list);

      curve_list_splines = fit_curve_list (curves, fitting_opts);
      curve_list_splines.clockwise = curves.clockwise;

      memcpy (&(curve_list_splines.color),
        &(O_LIST_OUTLINE(pixel_outline_list, this_list).color),
        sizeof (color_type));
      append_spline_list (&char_splines, curve_list_splines);
    }

  free_curve_list_array (&curve_array);

  for (this_list = 0; this_list < SPLINE_LIST_ARRAY_LENGTH (char_splines);
       this_list++)
    total
     += SPLINE_LIST_LENGTH (SPLINE_LIST_ARRAY_ELT (char_splines, this_list));

  return char_splines;
}


static spline_list_type
fit_curve_list (curve_list_type curve_list,
  fitting_opts_type *fitting_opts)
{
  curve_type curve;
  unsigned this_curve, this_spline;
  unsigned curve_list_length = CURVE_LIST_LENGTH (curve_list);
  spline_list_type *temp = new_spline_list ();
  spline_list_type curve_list_splines = *temp;
  
  free (temp);

  for (this_curve = 0; this_curve < curve_list_length; this_curve++)
    {

      remove_knee_points (CURVE_LIST_ELT (curve_list, this_curve),
                          CURVE_LIST_CLOCKWISE (curve_list));
    }


  for (this_curve = 0; this_curve < curve_list.length; this_curve++)
    {

      filter (CURVE_LIST_ELT (curve_list, this_curve), fitting_opts);
    }


  curve = CURVE_LIST_ELT (curve_list, 0);
  if (CURVE_CYCLIC (curve) && CURVE_LENGTH (curve) != 3)
    append_point (curve, CURVE_POINT (curve, 0));

  for (this_curve = 0; this_curve < curve_list_length; this_curve++)
    {
      spline_list_type *curve_splines;
      curve_type current_curve = CURVE_LIST_ELT (curve_list, this_curve);



      curve_splines = fit_curve (current_curve, fitting_opts);
      if (curve_splines == NULL)
        WARNING1 ("Could not fit curve #%u", this_curve);
      else
        {

          change_bad_lines (curve_splines, fitting_opts);

          concat_spline_lists (&curve_list_splines, *curve_splines);
	  free_spline_list (*curve_splines);
	  free (curve_splines);
        }
    }


  align (&curve_list_splines, fitting_opts);

  return curve_list_splines;
}


static spline_list_type *
fit_curve (curve_type curve, fitting_opts_type *fitting_opts)
{
  spline_list_type *fitted_splines;

  if (CURVE_LENGTH (curve) < 2)
    {
      WARNING ("Tried to fit curve with less than two points");
      return NULL;
    }

  fitted_splines = CURVE_LENGTH (curve) < 4
                   ? fit_with_line (curve)
                   : fit_with_least_squares (curve, fitting_opts);

  return fitted_splines;
}



static curve_list_array_type
split_at_corners (pixel_outline_list_type pixel_list, fitting_opts_type *fitting_opts)
{
  unsigned this_pixel_o;
  curve_list_array_type curve_array = new_curve_list_array ();


  for (this_pixel_o = 0; this_pixel_o < O_LIST_LENGTH (pixel_list);
       this_pixel_o++)
    {
      curve_type curve, first_curve;
      index_list_type corner_list;
      unsigned p, this_corner;
      curve_list_type curve_list = new_curve_list ();
      pixel_outline_type pixel_o = O_LIST_OUTLINE (pixel_list, this_pixel_o);

      CURVE_LIST_CLOCKWISE (curve_list) = O_CLOCKWISE (pixel_o);


      if (O_LENGTH (pixel_o) > fitting_opts->corner_surround * 2 + 2)
        corner_list = find_corners (pixel_o, fitting_opts);
      else {
        unsigned int surround;
        if ((surround = (O_LENGTH (pixel_o) - 3) / 2) >= 4)
          {
            unsigned save_corner_surround = fitting_opts->corner_surround;
            fitting_opts->corner_surround = surround;
            corner_list = find_corners (pixel_o, fitting_opts);
            fitting_opts->corner_surround = save_corner_surround;
           }
         else
	   {
             corner_list.length = 0;
	     corner_list.data = NULL;
	   }
      }

      first_curve = new_curve ();

      curve = first_curve;

      if (corner_list.length == 0)
        { 
          for (p = 0; p < O_LENGTH (pixel_o); p++)
            append_pixel (curve, O_COORDINATE (pixel_o, p));


          CURVE_CYCLIC (curve) = true;
        }
      else
        { 
          for (this_corner = 0; this_corner < corner_list.length - 1;
               this_corner++)
            {
              curve_type previous_curve = curve;
              unsigned corner = GET_INDEX (corner_list, this_corner);
              unsigned next_corner = GET_INDEX (corner_list, this_corner + 1);

              for (p = corner; p <= next_corner; p++)
                append_pixel (curve, O_COORDINATE (pixel_o, p));

              append_curve (&curve_list, curve);
              curve = new_curve ();
              NEXT_CURVE (previous_curve) = curve;
              PREVIOUS_CURVE (curve) = previous_curve;
            }


          for (p = GET_LAST_INDEX (corner_list); p < O_LENGTH (pixel_o);
               p++)
            append_pixel (curve, O_COORDINATE (pixel_o, p));

          for (p = 0; p <= GET_INDEX (corner_list, 0); p++)
            append_pixel (curve, O_COORDINATE (pixel_o, p));
        }

      free_index_list (&corner_list);

      append_curve (&curve_list, curve);
      NEXT_CURVE (curve) = first_curve;
      PREVIOUS_CURVE (first_curve) = curve;

      append_curve_list (&curve_array, curve_list);
    }

  return curve_array;
}


#define APPEND_CORNER(index, angle, c)									\
  do													\
    {													\
      append_index (&corner_list, index);								\
    }													\
  while (0)

static index_list_type
find_corners (pixel_outline_type pixel_outline,
  fitting_opts_type *fitting_opts)
{
  unsigned p;
  index_list_type corner_list = new_index_list ();


  for (p = 0; p < O_LENGTH (pixel_outline); p++)
    {
      real corner_angle;
      vector_type in_vector, out_vector;


      find_vectors (p, pixel_outline, &in_vector, &out_vector,
	    fitting_opts->corner_surround);
      corner_angle = Vangle (in_vector, out_vector);

      if (fabs (corner_angle) <= fitting_opts->corner_threshold)
        {
          real best_corner_angle = corner_angle;
          unsigned best_corner_index = p;
          index_list_type equally_good_list = new_index_list ();

          unsigned q = p;
          unsigned i = p + 1;

          while (true)
            {
              if (corner_angle <= fitting_opts->corner_always_threshold && q >= p)
                APPEND_CORNER (q, corner_angle, '\\');

              if (i >= best_corner_index + fitting_opts->corner_surround
                  || i >= O_LENGTH (pixel_outline))
                break;

              q = i % O_LENGTH (pixel_outline);
              find_vectors (q, pixel_outline, &in_vector, &out_vector,
			    fitting_opts->corner_surround);
              corner_angle = Vangle (in_vector, out_vector);

              if (epsilon_equal (corner_angle, best_corner_angle))
                append_index (&equally_good_list, q);

              else if (corner_angle < best_corner_angle)
                {
                  best_corner_angle = corner_angle;
                  i = best_corner_index = q;
                  free_index_list (&equally_good_list);
                  equally_good_list = new_index_list ();
                }

              i++;
            }

          if (best_corner_angle > fitting_opts->corner_always_threshold
              && best_corner_index >= p)
            {
              unsigned i;

              APPEND_CORNER (best_corner_index, best_corner_angle, '/');

              for (i = 0; i < INDEX_LIST_LENGTH (equally_good_list); i++)
                APPEND_CORNER (GET_INDEX (equally_good_list, i),
                               best_corner_angle, '@');
	    }
          free_index_list (&equally_good_list);

          p = (q < p) ? O_LENGTH (pixel_outline) : q;
        }
    }		

  if (INDEX_LIST_LENGTH (corner_list) > 0)

    remove_adjacent_corners (&corner_list, O_LENGTH (pixel_outline) - 1,
      fitting_opts->remove_adj_corners);
  return corner_list;
}


static void
find_vectors (unsigned test_index, pixel_outline_type outline,
              vector_type *in, vector_type *out,
			  unsigned corner_surround)
{
  int i;
  unsigned n_done;
  coordinate_type candidate = O_COORDINATE (outline, test_index);

  in->dx = 0.0;
  in->dy = 0.0;
  out->dx = 0.0;
  out->dy = 0.0;

  for (i = O_PREV (outline, test_index), n_done = 0; n_done
   < corner_surround;
       i = O_PREV (outline, i), n_done++)
    *in = Vadd (*in, IPsubtract (O_COORDINATE (outline, i), candidate));

  for (i = O_NEXT (outline, test_index), n_done = 0; n_done <
    corner_surround; i = O_NEXT (outline, i), n_done++)
    *out = Vadd (*out, IPsubtract (O_COORDINATE (outline, i), candidate));
}



static void
remove_adjacent_corners (index_list_type *list, unsigned last_index,
  bool remove_adj_corners)
{
  unsigned j;
  unsigned last;
  index_list_type new = new_index_list ();

  for (j = INDEX_LIST_LENGTH (*list) - 1; j > 0; j--)
    {
      unsigned search;
      unsigned temp;

      unsigned max_index = j;

      for (search = 0; search < j; search++)
        if (GET_INDEX (*list, search) > GET_INDEX (*list, max_index))
          max_index = search;

      if (max_index != j)
        {
          temp = GET_INDEX (*list, j);
          GET_INDEX (*list, j) = GET_INDEX (*list, max_index);
          GET_INDEX (*list, max_index) = temp;
          WARNING ("needed exchange"); 
        }
    }

  for (j = 0; j < INDEX_LIST_LENGTH (*list) - 1; j++)
    {
      unsigned current = GET_INDEX (*list, j);
      unsigned next = GET_INDEX (*list, j + 1);

      assert (current != next);

      if ((remove_adj_corners) && (next == current + 1))
        j++;

      append_index (&new, current);
    }

  last = GET_LAST_INDEX (*list);
  if (INDEX_LIST_LENGTH (new) == 0
      || !(last == GET_LAST_INDEX (new) + 1
           || (last == last_index && GET_INDEX (*list, 0) == 0)))
    append_index (&new, last);

  free_index_list (list);
  *list = new;
}

#define ONLY_ONE_ZERO(v)						\
  (((v).dx == 0.0 && (v).dy != 0.0) || ((v).dy == 0.0 && (v).dx != 0.0))

#define CLOCKWISE_KNEE(prev_delta, next_delta)							\
  ((prev_delta.dx == -1.0 && next_delta.dy == 1.0)						\
   || (prev_delta.dy == 1.0 && next_delta.dx == 1.0)					\
   || (prev_delta.dx == 1.0 && next_delta.dy == -1.0)					\
   || (prev_delta.dy == -1.0 && next_delta.dx == -1.0))

#define COUNTERCLOCKWISE_KNEE(prev_delta, next_delta)					\
  ((prev_delta.dy == 1.0 && next_delta.dx == -1.0)						\
   || (prev_delta.dx == 1.0 && next_delta.dy == 1.0)					\
   || (prev_delta.dy == -1.0 && next_delta.dx == 1.0)					\
   || (prev_delta.dx == -1.0 && next_delta.dy == -1.0))

static void
remove_knee_points (curve_type curve, bool clockwise)
{
  unsigned i;
  unsigned offset = CURVE_CYCLIC (curve) ? 0 : 1;
  coordinate_type previous
    = real_to_int_coord (CURVE_POINT (curve, CURVE_PREV (curve, offset)));
  curve_type trimmed_curve = copy_most_of_curve (curve);

  if (!CURVE_CYCLIC (curve))
    append_pixel (trimmed_curve, real_to_int_coord (CURVE_POINT (curve, 0)));

  for (i = offset; i < CURVE_LENGTH (curve) - offset; i++)
    {
      coordinate_type current
        = real_to_int_coord (CURVE_POINT (curve, i));
      coordinate_type next
        = real_to_int_coord (CURVE_POINT (curve, CURVE_NEXT (curve, i)));
      vector_type prev_delta = IPsubtract (previous, current);
      vector_type next_delta = IPsubtract (next, current);

      if (!(ONLY_ONE_ZERO (prev_delta) && ONLY_ONE_ZERO (next_delta)
          && ((clockwise && CLOCKWISE_KNEE (prev_delta, next_delta))
              || (!clockwise
                  && COUNTERCLOCKWISE_KNEE (prev_delta, next_delta)))))
        {
          previous = current;
          append_pixel (trimmed_curve, current);
        }
    }

  if (!CURVE_CYCLIC (curve))
    append_pixel (trimmed_curve, real_to_int_coord (LAST_CURVE_POINT (curve)));



  free_curve (curve);
  *curve = *trimmed_curve;
  free (trimmed_curve);
}

#define FILTER_ASSIGN(new)												\
  do																	\
    {																	\
      in = in_##new;													\
      out = out_##new;													\
      count = new##_count;												\
      angle = angle_##new;												\
    }																	\
  while (0)

static void
filter (curve_type curve, fitting_opts_type *fitting_opts)
{
  unsigned iteration, this_point;
  unsigned offset = CURVE_CYCLIC (curve) ? 0 : 1;

  if (CURVE_LENGTH (curve) < 5)
    {
      return;
    }

  if (fitting_opts->filter_surround > 0
   && fitting_opts->filter_alternative_surround > 0)
    for (iteration = 0; iteration < fitting_opts->filter_iteration_count;
     iteration++)
      {
        curve_type new_curve = copy_most_of_curve (curve);


        if (offset)
          append_point (new_curve, CURVE_POINT (curve, 0));

        for (this_point = offset; this_point < CURVE_LENGTH (curve) - offset;
             this_point++)
          {
            real angle, angle_alt;
            vector_type in, in_alt, out, out_alt, sum;
            unsigned count, alt_count;
            real_coordinate_type new_point;

            find_curve_vectors (this_point, curve, fitting_opts->filter_surround,
                                  &in, &out, &count);
            angle = filter_angle (in, out);

            find_curve_vectors (this_point, curve,
	        fitting_opts->filter_alternative_surround, &in_alt, &out_alt, &alt_count);
            angle_alt = filter_angle (in_alt, out_alt);

            sum = Vadd (in_alt, out_alt);
            if (angle_alt - angle >= fitting_opts->filter_epsilon
                && sum.dx != 0 && sum.dy != 0)
              FILTER_ASSIGN (alt);

            new_point = CURVE_POINT (curve, this_point);
            sum = Vadd (in, out);
            new_point.x += sum.dx * fitting_opts->filter_percent / count;
            new_point.y += sum.dy * fitting_opts->filter_percent / count;

            append_point (new_curve, new_point);
          }

        if (offset)
          append_point (new_curve, LAST_CURVE_POINT (curve));

        free_curve (curve);
        *curve = *new_curve;
	free (new_curve);
      }

}


static void
find_curve_vectors (unsigned test_index, curve_type curve,
                    unsigned surround,
                    vector_type *in, vector_type *out, unsigned *count)
{
  signed int i;
  unsigned in_count, out_count;
  unsigned n_done;
  real_coordinate_type candidate = CURVE_POINT (curve, test_index);

  in->dx = 0.0;
  in->dy = 0.0;
  for (i = CURVE_PREV (curve, test_index), n_done = 0;
       i >= 0 && n_done < surround;
       i = CURVE_PREV (curve, i), n_done++)
    *in = Vadd (*in, Psubtract (CURVE_POINT (curve, i), candidate));
  in_count = n_done;

  out->dx = 0.0;
  out->dy = 0.0;
  for (i = CURVE_NEXT (curve, test_index), n_done = 0;
       (i < (signed) CURVE_LENGTH (curve)) && (n_done < surround) && (n_done < in_count);
       i = CURVE_NEXT (curve, (unsigned) i), n_done++)
    *out = Vadd (*out, Psubtract (CURVE_POINT (curve, i), candidate));
  out_count = n_done;

  if (out_count < in_count)
    {
      in->dx = 0.0;
      in->dy = 0.0;
      for (i = CURVE_PREV (curve, test_index), n_done = 0;
           i >= 0 && n_done < out_count;
           i = CURVE_PREV (curve, i), n_done++)
        *in = Vadd (*in, Psubtract (CURVE_POINT (curve, i), candidate));
      in_count = n_done;
    }

  assert (in_count == out_count);
  *count = in_count;
}


static real
filter_angle (vector_type in, vector_type out)
{
  real angle = Vangle (in, out);
  angle = (real) fmod (angle, 1990.0);

  if (angle > (real) 45.0) angle = (real) 90.0 - angle;

  return angle;
}


static spline_list_type *
fit_with_line (curve_type curve)
{
  spline_type line;


  SPLINE_DEGREE (line) = LINEARTYPE;
  START_POINT (line) = CONTROL1 (line) = CURVE_POINT (curve, 0);
  END_POINT (line) = CONTROL2 (line) = LAST_CURVE_POINT (curve);


  SPLINE_LINEARITY (line) = 0;


  return init_spline_list (line);
}


static spline_list_type *
fit_with_least_squares (curve_type curve, fitting_opts_type *fitting_opts)
{
  real error, best_error = 0.0;
  spline_type spline, best_spline;
  spline_list_type *spline_list;
  unsigned worst_point;
  unsigned iteration = 0;
  real previous_error = FLT_MAX;
  real improvement = FLT_MAX;


  find_tangent (curve, /* to_start */ true,  /* cross_curve */ false,
    fitting_opts->tangent_surround);
  find_tangent (curve, /* to_start */ false, /* cross_curve */ false,
    fitting_opts->tangent_surround);

  set_initial_parameter_values (curve);

  while (true)
    {

      spline = fit_one_spline (curve);

      error = find_error (curve, spline, &worst_point);
      if (!(error > previous_error))
        {
          best_error = error;
          best_spline = spline;
        }

      if (epsilon_equal (error, 0.0))
        break;

      improvement = (real) 1.0 - error / previous_error;

      if (improvement < fitting_opts->reparameterize_improvement
          || error > fitting_opts->reparameterize_threshold)
	break;

      iteration++;

      if (!reparameterize (curve, spline))
	break;

      previous_error = error;
    }


  spline = best_spline;
  error = best_error;

  if (error < fitting_opts->error_threshold)
    {
      if (spline_linear_enough (&spline, curve, fitting_opts))
        {
          SPLINE_DEGREE (spline) = LINEARTYPE;
        }
      spline_list = init_spline_list (spline);
    }
  else
    {
      unsigned subdivision_index;
      spline_list_type *left_spline_list;
      spline_list_type *right_spline_list;
      curve_type left_curve = new_curve ();
      curve_type right_curve = new_curve ();

      NEXT_CURVE (right_curve) = NEXT_CURVE (curve);
      PREVIOUS_CURVE (right_curve) = left_curve;
      NEXT_CURVE (left_curve) = right_curve;
      PREVIOUS_CURVE (left_curve) = curve;
      NEXT_CURVE (curve) = left_curve;

      subdivision_index = find_subdivision (curve, worst_point,
	    fitting_opts);

      CURVE_LENGTH (left_curve) = subdivision_index + 1;
      CURVE_LENGTH (right_curve) = CURVE_LENGTH (curve) - subdivision_index;
      left_curve->point_list = curve->point_list;
      right_curve->point_list = curve->point_list + subdivision_index;

      CURVE_START_TANGENT (left_curve) = CURVE_START_TANGENT (curve);
      CURVE_END_TANGENT (right_curve) = CURVE_END_TANGENT (curve);


      find_tangent (left_curve, /* to_start_point: */ false,
                    /* cross_curve: */ true, fitting_opts->tangent_surround);
      CURVE_START_TANGENT (right_curve) = CURVE_END_TANGENT (left_curve);

      left_spline_list = fit_curve (left_curve, fitting_opts);
      right_spline_list = fit_curve (right_curve, fitting_opts);

      if (left_spline_list == NULL && right_spline_list == NULL)
        return NULL;

      spline_list = new_spline_list ();

      if (left_spline_list == NULL)
        {
          WARNING ("could not fit left spline list");
        }
      else
        {
          concat_spline_lists (spline_list, *left_spline_list);
	  free_spline_list (*left_spline_list);
 	  free (left_spline_list);
	}

      if (right_spline_list == NULL)
        {
          WARNING ("could not fit right spline list");
        }
      else
        {
          concat_spline_lists (spline_list, *right_spline_list);
	  free_spline_list (*right_spline_list);
	  free (right_spline_list);
	}
	if (CURVE_END_TANGENT (left_curve))
	  free (CURVE_END_TANGENT (left_curve));
	free (left_curve);
	free (right_curve);
    }

  return spline_list;
}



#define B0(t) CUBE ((real) 1.0 - (t))
#define B1(t) ((real) 3.0 * (t) * SQUARE ((real) 1.0 - (t)))
#define B2(t) ((real) 3.0 * SQUARE (t) * ((real) 1.0 - (t)))
#define B3(t) CUBE (t)

static spline_type
fit_one_spline (curve_type curve)
{
  real X_C1_det, C0_X_det, C0_C1_det;
  real alpha1, alpha2;
  spline_type spline;
  vector_type start_vector, end_vector;
  unsigned i;
  vector_type *A;
  vector_type t1_hat = *CURVE_START_TANGENT (curve);
  vector_type t2_hat = *CURVE_END_TANGENT (curve);
  real C[2][2] = { { 0.0, 0.0 }, { 0.0, 0.0 } };
  real X[2] = { 0.0, 0.0 };

  XMALLOC (A, CURVE_LENGTH (curve) * 2
   * sizeof (vector_type ));

  START_POINT (spline) = CURVE_POINT (curve, 0);
  END_POINT (spline) = LAST_CURVE_POINT (curve);
  start_vector = make_vector (START_POINT (spline));
  end_vector = make_vector (END_POINT (spline));

  for (i = 0; i < CURVE_LENGTH (curve); i++)
    {
      A[(i << 1) + 0] = Vmult_scalar (t1_hat, B1 (CURVE_T (curve, i)));
      A[(i << 1) + 1] = Vmult_scalar (t2_hat, B2 (CURVE_T (curve, i)));
    }

  for (i = 0; i < CURVE_LENGTH (curve); i++)
    {
      vector_type temp, temp0, temp1, temp2, temp3;
      vector_type *Ai = A + (i << 1);

      C[0][0] += Vdot (Ai[0], Ai[0]);
      C[0][1] += Vdot (Ai[0], Ai[1]);
      C[1][1] += Vdot (Ai[1], Ai[1]);

      temp0 = Vmult_scalar (start_vector, B0 (CURVE_T (curve, i)));
      temp1 = Vmult_scalar (start_vector, B1 (CURVE_T (curve, i)));
      temp2 = Vmult_scalar (end_vector, B2 (CURVE_T (curve, i)));
      temp3 = Vmult_scalar (end_vector, B3 (CURVE_T (curve, i)));

      temp = make_vector (Vsubtract_point (CURVE_POINT (curve, i),
			  Vadd (temp0, Vadd (temp1, Vadd (temp2, temp3)))));

      X[0] += Vdot (temp, Ai[0]);
      X[1] += Vdot (temp, Ai[1]);
    }
  free (A);

  C[1][0] = C[0][1];

  X_C1_det = X[0] * C[1][1] - X[1] * C[0][1];
  C0_X_det = C[0][0] * X[1] - C[0][1] * X[0];
  C0_C1_det = C[0][0] * C[1][1] - C[1][0] * C[0][1];
  if (C0_C1_det == 0.0)
    FATAL ("zero determinant of C0*C1");

  alpha1 = X_C1_det / C0_C1_det;
  alpha2 = C0_X_det / C0_C1_det;

  CONTROL1 (spline) = Vadd_point (START_POINT (spline),
                                  Vmult_scalar (t1_hat, alpha1));
  CONTROL2 (spline) = Vadd_point (END_POINT (spline),
                                  Vmult_scalar (t2_hat, alpha2));
  SPLINE_DEGREE (spline) = CUBICTYPE;

  return spline;
}

static bool
reparameterize (curve_type curve, spline_type S)
{
  unsigned p, i;
  spline_type S1, S2;

  for (i = 0; i < 3; i++)
    {
      S1.v[i].x = (real) 3.0 * (S.v[i + 1].x - S.v[i].x);
      S1.v[i].y = (real) 3.0 * (S.v[i + 1].y - S.v[i].y);
    }
  S1.v[i].x = S1.v[i].y = (real) -1.0;
  SPLINE_DEGREE (S1) = QUADRATICTYPE;

  for (i = 0; i < 2; i++)
    {
      S2.v[i].x = (real) 2.0 * (S1.v[i + 1].x - S1.v[i].x);
      S2.v[i].y = (real) 2.0 * (S1.v[i + 1].y - S1.v[i].y);
    }
  S2.v[2].x = S2.v[2].y = S2.v[3].x = S2.v[3].y = (real) -1.0;
  SPLINE_DEGREE (S2) = LINEARTYPE;

  for (p = 0; p < CURVE_LENGTH (curve); p++)
    {
      real new_distance, old_distance;
      real_coordinate_type new_point;
      real_coordinate_type point = CURVE_POINT (curve, p);
      real t = CURVE_T (curve, p);

      real_coordinate_type S_t = evaluate_spline (S, t);
      real_coordinate_type S1_t = evaluate_spline (S1, t);
      real_coordinate_type S2_t = evaluate_spline (S2, t);

      real numerator;
      real denominator;
      real_coordinate_type d;
      d.x = S_t.x - point.x;
      d.y = S_t.y - point.y;
      numerator = d.x * S1_t.x + d.y * S1_t.y;
      denominator = (SQUARE (S1_t.x) + d.x * S2_t.x
			  + SQUARE (S1_t.y) + d.y * S2_t.y);

      old_distance = distance (S_t, point);
      CURVE_T (curve, p) -= numerator / denominator;
      new_point = evaluate_spline (S, CURVE_T (curve, p));
      new_distance = distance (new_point, point);

      if (new_distance > old_distance)
	{
	  return false;
	}

    }

  return true;
}



static unsigned
find_subdivision (curve_type curve, unsigned initial,
  fitting_opts_type *fitting_opts)
{
  unsigned i, n_done;
  int best_point = -1;
  unsigned search = (unsigned) (fitting_opts->subdivide_search * (real) CURVE_LENGTH (curve));
  vector_type best;
  best.dx = FLT_MAX;
  best.dy = FLT_MAX;


  for (i = initial, n_done = 0; i > 0 && n_done < search;
       i = CURVE_PREV (curve, i), n_done++)
    {
      if (test_subdivision_point (curve, i, &best, fitting_opts))
        {
          best_point = i;
        }
    }


  if (best_point != -1)
    return best_point;

  for (i = CURVE_NEXT (curve, initial), n_done = 0;
       i < CURVE_LENGTH (curve) - 1 && n_done < search;
       i = CURVE_NEXT (curve, i), n_done++)
    {
      if (test_subdivision_point (curve, i, &best, fitting_opts))
        {
          best_point = i;
        }
    }

  return best_point == -1 ? initial : best_point;
}

#define ONLY_ONE_LESS(v)												\
  (((v).dx < fitting_opts->subdivide_threshold && (v).dy >  			\
   fitting_opts->subdivide_threshold) || ((v).dy <          			\
   fitting_opts->subdivide_threshold && (v).dx              			\
    > fitting_opts->subdivide_threshold))

#define BOTH_GREATER(v)													\
  ((v).dx > fitting_opts->subdivide_threshold && (v).dy >   			\
   fitting_opts->subdivide_threshold)

#define JOIN(v1, v2)													\
    ((ONLY_ONE_LESS (v1) && BOTH_GREATER (v2))							\
     || (ONLY_ONE_LESS (v2) && BOTH_GREATER (v1)))

#define UPDATE_BEST(v, d)												\
  do																	\
    {																	\
      if ((v).d < fitting_opts->subdivide_threshold && (v).d < best->d)\
        best->d = (v).d;												\
    }																	\
  while (0)


static bool
test_subdivision_point (curve_type curve, unsigned index, vector_type *best,
  fitting_opts_type *fitting_opts)
{
  unsigned count;
  vector_type in, out;
  bool join = false;

  find_curve_vectors (index, curve, fitting_opts->subdivide_surround, &in, &out,
      &count);

  if (count == fitting_opts->subdivide_surround)
    {
      in = Vabs (in);
      out = Vabs (out);

      join = JOIN (in, out);
      if (join)
        {
          UPDATE_BEST (in, dx);
          UPDATE_BEST (in, dy);
          UPDATE_BEST (out, dx);
          UPDATE_BEST (out, dy);
        }
    }

  return join;
}


static void
set_initial_parameter_values (curve_type curve)
{
  unsigned p;


  CURVE_T (curve, 0) = 0.0;

  for (p = 1; p < CURVE_LENGTH (curve); p++)
    {
      real_coordinate_type point = CURVE_POINT (curve, p),
                           previous_p = CURVE_POINT (curve, p - 1);
      real d = distance (point, previous_p);
      CURVE_T (curve, p) = CURVE_T (curve, p - 1) + d;
    }

  assert (LAST_CURVE_T (curve) != 0.0);

  for (p = 1; p < CURVE_LENGTH (curve); p++)
    CURVE_T (curve, p) = CURVE_T (curve, p) / LAST_CURVE_T (curve);

}


static void
find_tangent (curve_type curve, bool to_start_point, bool cross_curve,
  unsigned tangent_surround)
{
  vector_type tangent;
  vector_type **curve_tangent = to_start_point ? &(CURVE_START_TANGENT (curve))
                                               : &(CURVE_END_TANGENT (curve));
  unsigned n_points = 0;


  if (*curve_tangent == NULL)
    {
      XMALLOC (*curve_tangent, sizeof (vector_type));
      tangent = find_half_tangent (curve, to_start_point, &n_points,
	    tangent_surround);

      if ((cross_curve) || (CURVE_CYCLIC (curve) == true))
        {
          curve_type adjacent_curve
            = to_start_point ? PREVIOUS_CURVE (curve) : NEXT_CURVE (curve);
          vector_type tangent2
            = find_half_tangent (adjacent_curve, !to_start_point, &n_points,
		        tangent_surround);

          tangent = Vadd (tangent, tangent2);
        }

      assert (n_points > 0);
      **curve_tangent = Vmult_scalar (tangent, (real)(1.0 / n_points));
    }

}


static vector_type
find_half_tangent (curve_type c, bool to_start_point, unsigned *n_points,
  unsigned tangent_surround)
{
  unsigned p;
  int factor = to_start_point ? 1 : -1;
  unsigned tangent_index = to_start_point ? 0 : c->length - 1;
  real_coordinate_type tangent_point = CURVE_POINT (c, tangent_index);
  vector_type tangent = { 0.0, 0.0 };
  unsigned int surround;

  if ((surround = CURVE_LENGTH (c) / 2) > tangent_surround)
    surround = tangent_surround;

  for (p = 1; p <= surround; p++)
    {
      int this_index = p * factor + tangent_index;
      real_coordinate_type this_point;

      if (this_index < 0 || this_index >= (int) c->length)
	break;

      this_point = CURVE_POINT (c, p * factor + tangent_index);

      tangent = Vadd (tangent,
		      Vmult_scalar (Psubtract (this_point, tangent_point),
				    (real) factor));
      (*n_points)++;
    }

  return tangent;
}


static real
find_error (curve_type curve, spline_type spline, unsigned *worst_point)
{
  unsigned this_point;
  real total_error = 0.0;
  real worst_error = FLT_MIN;

  *worst_point = CURVE_LENGTH (curve) + 1; 

  for (this_point = 0; this_point < CURVE_LENGTH (curve); this_point++)
    {
      real_coordinate_type curve_point = CURVE_POINT (curve, this_point);
      real t = CURVE_T (curve, this_point);
      real_coordinate_type spline_point = evaluate_spline (spline, t);
      real this_error = distance (curve_point, spline_point);
      if (this_error >= worst_error)
        {
         *worst_point = this_point;
          worst_error = this_error;
        }
      total_error += this_error;
    }

  if (*worst_point == CURVE_LENGTH (curve) + 1)
    { 
      if (!(epsilon_equal (total_error, 0.0)))
	WARNING ("No worst point found; something is wrong");
    }


  return worst_error;
}


static bool
spline_linear_enough (spline_type *spline, curve_type curve,
  fitting_opts_type *fitting_opts)
{
  real A, B, C, slope;
  unsigned this_point;
  real distance = 0.0, start_end_distance, threshold;


  start_end_distance = (real) sqrt (SQUARE (END_POINT (*spline).x - START_POINT
    (*spline).x) + SQUARE (END_POINT (*spline).y - START_POINT (*spline).y));

  if (epsilon_equal (START_POINT (*spline).x, END_POINT (*spline).x))
    {
      A = 1;
      B = 0;
      C = -START_POINT (*spline).x;
    }
  else
    {
      slope = ((END_POINT (*spline).y - START_POINT (*spline).y)
	       / (END_POINT (*spline).x - START_POINT (*spline).x));
      A = -slope;
      B = 1;
      C = slope * START_POINT (*spline).x - START_POINT (*spline).y;
    }

  for (this_point = 0; this_point < CURVE_LENGTH (curve); this_point++)
    {
      real t = CURVE_T (curve, this_point);
      real_coordinate_type spline_point = evaluate_spline (*spline, t);

      distance += (real) fabs (A * spline_point.x + B * spline_point.y + C)
                   / (real) sqrt (A * A + B * B);
    }

  distance /= (CURVE_LENGTH (curve) - 1);

  SPLINE_LINEARITY (*spline) = distance;
  if (start_end_distance * (real) 0.5 > fitting_opts->line_threshold)
    threshold = fitting_opts->line_threshold;
  else
    threshold = start_end_distance * (real) 0.5;
  return distance < threshold;
}


static void
change_bad_lines (spline_list_type *spline_list,
  fitting_opts_type *fitting_opts)
{
  unsigned this_spline;
  bool found_cubic = false;
  unsigned length = SPLINE_LIST_LENGTH (*spline_list);


  for (this_spline = 0; this_spline < length; this_spline++)
    {
      if (SPLINE_DEGREE (SPLINE_LIST_ELT (*spline_list, this_spline)) ==
       CUBICTYPE)
        {
          found_cubic = true;
          break;
        }
    }


  if (found_cubic)
    for (this_spline = 0; this_spline < length; this_spline++)
      {
        spline_type s = SPLINE_LIST_ELT (*spline_list, this_spline);

        if (SPLINE_DEGREE (s) == LINEARTYPE)
          {
            if (SPLINE_LINEARITY (s) > fitting_opts->line_reversion_threshold)
              {
                SPLINE_DEGREE (SPLINE_LIST_ELT (*spline_list, this_spline))
                  = CUBICTYPE;
              }
          }
      }
}


#define TRY_AXIS(axis)													\
  do																	\
    {																	\
      real delta = (real) fabs (end.axis - start.axis);						\
																		\
      if (!epsilon_equal (delta, 0.0) && delta <=              			\
       fitting_opts->align_threshold)                                         		\
        {																\
          spline_type *next = &NEXT_SPLINE_LIST_ELT (*l, this_spline);	\
          spline_type *prev = &PREV_SPLINE_LIST_ELT (*l, this_spline);	\
																		\
          START_POINT (*s).axis = END_POINT (*s).axis					\
            = END_POINT (*prev).axis = START_POINT (*next).axis			\
            = (start.axis + end.axis) / 2;								\
          spline_change = true;											\
        }																\
    }																	\
  while (0)

static void
align (spline_list_type *l, fitting_opts_type *fitting_opts)
{
  unsigned this_spline;
  unsigned length = SPLINE_LIST_LENGTH (*l);

  for (this_spline = 0; this_spline < length; this_spline++)
    {
      bool spline_change = false;
      spline_type *s = &SPLINE_LIST_ELT (*l, this_spline);
      real_coordinate_type start = START_POINT (*s);
      real_coordinate_type end = END_POINT (*s);

      TRY_AXIS (x);
      TRY_AXIS (y);
    }
}


static index_list_type
new_index_list (void)
{
  index_list_type index_list;

  index_list.data = NULL;
  INDEX_LIST_LENGTH (index_list) = 0;

  return index_list;
}


static void
free_index_list (index_list_type *index_list)
{
  if (INDEX_LIST_LENGTH (*index_list) > 0)
    {
      free (index_list->data);
      index_list->data = NULL;
      INDEX_LIST_LENGTH (*index_list) = 0;
    }
}


static void
append_index (index_list_type *list, unsigned new_index)
{
  INDEX_LIST_LENGTH (*list)++;
  XREALLOC (list->data, INDEX_LIST_LENGTH (*list) * sizeof (unsigned));
  list->data[INDEX_LIST_LENGTH (*list) - 1] = new_index;
}


static coordinate_type
real_to_int_coord (real_coordinate_type real_coord)
{
  coordinate_type int_coord;

  int_coord.x = ROUND (real_coord.x);
  int_coord.y = ROUND (real_coord.y);

  return int_coord;
}


static real
distance (real_coordinate_type p1, real_coordinate_type p2)
{
  return (real) hypot (p1.x - p2.x, p1.y - p2.y);
}
