#include "message.h"
#include "types.h"
#include "spline.h"
#include "vector.h"
#include "xmem.h"
#include <assert.h>


void
print_spline (FILE *f, spline_type s)
{
  if (SPLINE_DEGREE (s) == LINEARTYPE)
    fprintf (f, "(%.3f,%.3f)--(%.3f,%.3f).\n",
                START_POINT (s).x, START_POINT (s).y,
                END_POINT (s).x, END_POINT (s).y);

  else if (SPLINE_DEGREE (s) == CUBICTYPE)
    fprintf (f, "(%.3f,%.3f)..ctrls(%.3f,%.3f)&(%.3f,%.3f)..(%.3f,%.3f).\n",
                START_POINT (s).x, START_POINT (s).y,
                CONTROL1 (s).x, CONTROL1 (s).y,
                CONTROL2 (s).x, CONTROL2 (s).y,
                END_POINT (s).x, END_POINT (s).y);

  else
    FATAL1 ("print_spline: strange degree (%d)", SPLINE_DEGREE (s));
}



real_coordinate_type
evaluate_spline (spline_type s, real t)
{
  spline_type V[4]; 
  signed i, j;
  real one_minus_t = (real) 1.0 - t;
  polynomial_degree degree = SPLINE_DEGREE (s);

  for (i = 0; i <= degree; i++)
    V[0].v[i] = s.v[i];

  for (j = 1; j <= degree; j++)
    for (i = 0; i <= degree - j; i++)
      {
        real_coordinate_type t1 = Pmult_scalar (V[j - 1].v[i], one_minus_t);
        real_coordinate_type t2 = Pmult_scalar (V[j - 1].v[i + 1], t);
        V[j].v[i] = Padd (t1, t2);
      }

  return V[degree].v[0];
}



spline_list_type *
new_spline_list (void)
{
  spline_list_type *answer;

  XMALLOC (answer, sizeof (spline_list_type));
  SPLINE_LIST_DATA (*answer) = NULL;
  SPLINE_LIST_LENGTH (*answer) = 0;

  return answer;
}



spline_list_type *
init_spline_list (spline_type spline)
{
  spline_list_type *answer;

  XMALLOC (answer, sizeof (spline_list_type));
  XMALLOC (SPLINE_LIST_DATA (*answer), sizeof (spline_type));
  SPLINE_LIST_ELT (*answer, 0) = spline;
  SPLINE_LIST_LENGTH (*answer) = 1;

  return answer;
}


void
free_spline_list (spline_list_type spline_list)
{
  if (SPLINE_LIST_DATA (spline_list) != NULL)
    free (SPLINE_LIST_DATA (spline_list));
}


void
append_spline (spline_list_type *l, spline_type s)
{
  assert (l != NULL);

  SPLINE_LIST_LENGTH (*l)++;
  XREALLOC (SPLINE_LIST_DATA (*l), SPLINE_LIST_LENGTH (*l) * sizeof (spline_type));
  LAST_SPLINE_LIST_ELT (*l) = s;
}


void
concat_spline_lists (spline_list_type *s1, spline_list_type s2)
{
  unsigned this_spline;
  unsigned new_length;

  assert (s1 != NULL);

  new_length = SPLINE_LIST_LENGTH (*s1) + SPLINE_LIST_LENGTH (s2);

  XREALLOC (SPLINE_LIST_DATA (*s1), new_length * sizeof (spline_type));

  for (this_spline = 0; this_spline < SPLINE_LIST_LENGTH (s2); this_spline++)
    SPLINE_LIST_ELT (*s1, SPLINE_LIST_LENGTH (*s1)++)
      = SPLINE_LIST_ELT (s2, this_spline);
}



spline_list_array_type
new_spline_list_array (void)
{
  spline_list_array_type answer;

  SPLINE_LIST_ARRAY_DATA (answer) = NULL;
  SPLINE_LIST_ARRAY_LENGTH (answer) = 0;

  return answer;
}



void
free_spline_list_array (spline_list_array_type *spline_list_array)
{
  unsigned this_list;

  for (this_list = 0;
       this_list < SPLINE_LIST_ARRAY_LENGTH (*spline_list_array);
       this_list++)
    free_spline_list (SPLINE_LIST_ARRAY_ELT (*spline_list_array, this_list));

  if (SPLINE_LIST_ARRAY_DATA (*spline_list_array) != NULL)
    free (SPLINE_LIST_ARRAY_DATA (*spline_list_array));

}


void
append_spline_list (spline_list_array_type *l, spline_list_type s)
{
  SPLINE_LIST_ARRAY_LENGTH (*l)++;
  XREALLOC (SPLINE_LIST_ARRAY_DATA (*l), SPLINE_LIST_ARRAY_LENGTH (*l) * sizeof (spline_list_type));
  LAST_SPLINE_LIST_ARRAY_ELT (*l) = s;
}

